import datetime
import os

import argparse
import configparser
import fnmatch
from pathlib import Path
import shutil
import webbrowser

from sarcharts.lib.chartjs import ChartJS
from sarcharts.lib.events import Events
from sarcharts.lib.metrics import Metrics
from sarcharts.lib.sadf import Sadf
from sarcharts.lib import util


class SarCharts:
    args = ()
    cwd = os.getcwd()
    default_cfg = f"{Path.home()}/.sarcharts.cfg"

    def valid_date(self, d):
        valid = util.valid_date_formats
        for v in valid:
            try:
                return datetime.datetime.strptime(d, v)
            except ValueError:
                pass
        raise argparse.ArgumentTypeError(
            f"not a valid date: {d!r}. Valid formats: {str(valid)}")

    def valid_config_path(self, path):
        if path == self.default_cfg and not os.path.exists(self.default_cfg):
            currpath = os.path.dirname(os.path.realpath(__file__))
            shutil.copyfile(
                f"{currpath}/conf/sarcharts.cfg-dist",
                self.default_cfg)
            return path
        if os.path.exists(path):
            return path
        else:
            raise argparse.ArgumentTypeError(f"not a valid path: {path!r}")

    def valid_path(self, path):
        if os.path.exists(path):
            return path
        else:
            raise argparse.ArgumentTypeError(f"not a valid path: {path!r}")

    def default_sarfiles(self):
        files = []
        for f in os.listdir(self.cwd):
            if fnmatch.fnmatch(f, 'sa??'):
                files.append(f"{self.cwd}/{str(f)}")
        return files

    def __init__(self):
        self.parser = argparse.ArgumentParser(
            description="SarCharts gets \"sysstat\" files from provided"
            + " `sarfilespaths` and generates dynamic HTML Charts."
            )
        self.parser.add_argument(
            '-c',
            '--configfile',
            help='Use alternative config file. Default `~/.sarcharts.cfg`.',
            default=self.default_cfg,
            type=self.valid_config_path
            )
        self.parser.add_argument(
            '-d',
            '--debug',
            help='Set debug level. Default `W`.',
            default='W',
            choices=['D', 'I', 'W', 'E']
            )
        self.parser.add_argument(
            '-e',
            '--eventfile',
            help='Add events csv file. Header: '
                 + '# date;hostname;eventname;eventdescription',
            type=self.valid_path
            )
        self.parser.add_argument(
            '-f',
            '--fromdate',
            help='Include metrics/events from this date.',
            default=datetime.datetime.strptime('1970-01-01', '%Y-%m-%d'),
            type=self.valid_date
            )
        self.parser.add_argument(
            '-m',
            '--metricfile',
            help='Add metrics csv file. Header: '
                 + '# date;hostname;metricname;metricvalue',
            type=self.valid_path
            )
        self.parser.add_argument(
            '-o',
            '--outputpath',
            help='Path to put output files. Default `./sarcharts`.',
            default='.'
            )
        self.parser.add_argument(
            '-t',
            '--todate',
            help='Include metrics/events before this date.',
            default=datetime.datetime.strptime('2039-01-01', '%Y-%m-%d'),
            type=self.valid_date
            )
        self.parser.add_argument(
            '-q',
            '--quiet',
            help="Don't show progress.",
            default=False,
            action='store_true'
            )
        self.parser.add_argument(
            'sarfilespaths',
            help='`sa` file/s to parse. Default: `./sa??`.',
            default=self.default_sarfiles(),
            type=self.valid_path,
            nargs='*'
            )
        self.args = self.parser.parse_args()

        # create required files on outputpath
        self.args.outputpath = self.args.outputpath + "/sarcharts"
        if os.path.exists(self.args.outputpath):
            util.debug(self.args, 'E',
                       f"Output path '{self.args.outputpath}' already exists.")
            # shutil.rmtree(self.args.outputpath)
        os.makedirs(self.args.outputpath + "/csvfiles")
        shutil.copytree(os.path.dirname(
                        os.path.realpath(__file__)) + "/html",
                        self.args.outputpath + "/html")

    def read_config(self):
        self.config = configparser.ConfigParser()
        self.config.read(self.args.configfile)

    def main(self):
        self.read_config()
        hidden_metrics = self.config.get('main', 'hidden_metrics').split("\n")
        hidden_custom = self.config.get('main', 'hidden_custom').split("\n")
        if len(self.args.sarfilespaths) > 0:
            sarfiles = util.get_filelist(self.args.sarfilespaths)
            if len(sarfiles) > 0:
                util.debug(self.args, 'D', "sarfiles: " + str(sarfiles))
                charts = Sadf().sar_to_chartjs(self.args, sarfiles)
                charts = Events.getCSVdata(self.args, charts)
                charts = Metrics.getCSVdata(self.args, charts)
                ChartJS(hidden_metrics,
                        hidden_custom).write_files(self.args, charts)
                util.debug(self.args, '', "Open SarCharts in default browser.")
                webbrowser.open(self.args.outputpath, 0, True)
            else:
                self.parser.print_help()
                util.debug(
                    self.args, 'E',
                    f"No valid `sa??` files in `{self.args.sarfilespaths}`.")



import datetime
import os
import re

import jinja2
from random import randrange

from sarcharts.lib.progressbar import ProgressBar


class ChartJS:

    def __init__(self, hidden_metrics, hidden_custom):
        self.hidden_metrics = hidden_metrics
        self.hidden_custom = hidden_custom

    def get_color_palette(self, datasets):
        number = 0
        for k, data in datasets.items():
            size = len(data)
            if size > number:
                number = size
        colors = []
        for i in range(0, number):
            r = randrange(225)
            g = randrange(225)
            b = randrange(225)
            colors.append(f"{r}, {g}, {b}")
        return colors

    def is_hidden_metric(self, string):
        for r in self.hidden_metrics:
            if r != "" and re.search(f'^{r}$', string):
                return True

    def is_hidden_custom(self, string):
        for r in self.hidden_custom:
            if r != "" and re.search(r, string):
                return True

    def write_files(self, args, charts):
        pb = ProgressBar()
        all_entries = 0
        for nodename in charts.keys():
            all_entries += len(charts[nodename]['activities'].keys())
        pb.quiet = args.quiet
        pb.all_entries = all_entries
        pb.start_time = datetime.datetime.now()
        pbi = 0
        # show full metric list
        # for activity, csvdata in charts[list(charts.keys())[0]]['activities'].items():
        #     for i in list(csvdata['datasets'].values())[0]:
        #         print(f"{activity}:.*:{i['label']}")
        #     break

        # write output
        for nodename, nodecharts in charts.items():
            for activity, csvdata in nodecharts['activities'].items():
                palette = self.get_color_palette(csvdata['datasets'])
                outputfile = args.outputpath + f"/{nodename}_{activity}.html"
                pbi += 1
                pb.print_bar(pbi, f"{activity}.")
                context = {
                    "chart": activity,
                    "xlabels": sorted(nodecharts['xlabels']),
                    "sysname": nodecharts['sysname'],
                    "release": nodecharts['release'],
                    "machine": nodecharts['machine'],
                    "numberofcpus": nodecharts['number-of-cpus'],
                    "timezone": nodecharts['timezone'],
                    "details": csvdata,
                    "events": nodecharts['events'],
                    "pages": sorted(nodecharts['activities'].keys()),
                    "colors": palette,
                    "hostname": nodename,
                    "hostnames": charts.keys()                    }
                parent = (
                    os.path.dirname(
                        os.path.realpath(__file__)) + "/../templates/"
                        )
                environment = (
                    jinja2.Environment(
                        loader=jinja2.FileSystemLoader(parent),
                        extensions=['jinja2.ext.loopcontrols'])
                )
                environment.filters["is_hidden_metric"] = self.is_hidden_metric
                environment.filters["is_hidden_custom"] = self.is_hidden_custom
                template = environment.get_template("chart.html")
                with open(outputfile, mode="w", encoding="utf-8") as results:
                    results.write(template.render(context))
        pb.finish(f"  Write Charts and CSV file to {args.outputpath}.")



import csv

from sarcharts.lib import util


class Metrics:

    def getCSVdata(args, charts):
        # date;hostname;metric1;metric2;...
        if args.metricfile:
            metrics = {}
            with open(args.metricfile) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for r, row in enumerate(csv_reader):
                    if r == 0:
                        for i in range(2, len(row)):
                            metrics[row[i]] = []
                    else:
                        if row[1] not in charts.keys():
                            util.debug(
                                args, 'W',
                                f"Host '{row[1]}' from '{args.metricfile}'"
                                + " not in 'sar' data. Metric ignored!!!")
                            continue
                        if util.in_date_range(args, row[0]):
                            charts[row[1]]['xlabels'].append(row[0])
                            for i in range(len(metrics.keys())):
                                k = list(metrics.keys())[i]
                                metrics[k].append({'x': row[0], 'y': row[i+2]})

            for a in charts[row[1]]['activities']:
                for d in charts[row[1]]['activities'][a]['datasets']:
                    for m in reversed(range(len(metrics.keys()))):
                        metric = list(metrics.keys())[m]
                        values = list(metrics.values())[m]
                        charts[row[1]]['activities'][a]['datasets'][d].insert(0, {
                            "label": f"extrametric_{metric}",
                            "yAxisID": f"y{m}",
                            "values": values
                            })

        return charts



import csv

from sarcharts.lib import util


class Events:

    def getCSVdata(args, charts):
        if args.eventfile:
            with open(args.eventfile) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for row in csv_reader:
                    if not row[0].startswith("#"):
                        if row[1] not in charts.keys():
                            util.debug(
                                args, 'W',
                                f"Host '{row[1]}' from '{args.eventfile}'"
                                + " not in 'sar' data.")
                            continue
                        if util.in_date_range(args, row[0]):
                            charts[row[1]]['xlabels'].append(row[0])
                            if row[2] not in charts[row[1]]['events']:
                                charts[row[1]]['events'][row[2]] = []
                            charts[row[1]]['events'][row[2]].append(
                                   {'date': util.valid_date(args, row[0]),
                                    'description': (
                                        row[3]
                                        .replace(r'"', r'\"')
                                        .replace(r'(', r'\(')
                                        .replace(r')', r'\)'))
                                    })
        return charts



import datetime
import json

from sarcharts.lib.progressbar import ProgressBar
from sarcharts.lib import util


class Sadf:

    def sar_to_json(self, args, inputfile, arg,):
        command = f"sadf -tj {inputfile} -- {arg}"
        [stdout, stderr] = util.exec_command(args, command)
        if stderr:
            if "Try to convert it to current format" in stderr:
                # tf = tempfile.NamedTemporaryFile(prefix="sarcharts")
                command = f"sadf -c {inputfile} > /tmp/sarcharts.tmp"
                [stdout, stderr] = util.exec_command(args, command)
                return self.sar_to_json(args, "/tmp/sarcharts.tmp", arg)
            elif "Requested activities not available" in stderr:
                util.debug(args, 'I', stderr.strip())
            else:
                util.debug(args, 'W', f"{command}\n    {stderr.strip()}")
        else:
            return json.loads(stdout)

    def merge_sarfiles(self, args, sarfiles):
        data = []
        for inputfile in sarfiles:
            jdata = self.sar_to_json(args, inputfile, '-A')
            if jdata:
                data.append(jdata)
            else:
                util.debug(args, 'W', f"Can't add data from {inputfile}.")
        return data

    def data_normalization(self, act, data):
        if isinstance(data, list) and isinstance(data[0], dict):
            print(f"{act} is type 1")
            headers = ";".join(data[0].keys())
            print(headers)

    def sar_to_chartjs(self, args, sarfiles):
        data = self.merge_sarfiles(args, sarfiles)
        linehead = "# hostname;interval;timestamp"
        pb = ProgressBar()
        pb.quiet = args.quiet
        all_entries = 0
        for idata in range(len(data)):
            for ihost in range(len(data[idata]['sysstat']['hosts'])):
                for istats in range(len(data[idata]['sysstat'][
                        'hosts'][ihost]['statistics'])):
                    all_entries += len(data[idata]['sysstat']['hosts'][
                        ihost]['statistics'][istats].keys())
        pb.all_entries = all_entries

        pb.start_time = datetime.datetime.now()
        pbi = 0
        charts = {}
        for idata in range(len(data)):
            hostsdata = data[idata]['sysstat']['hosts']
            for ihost in range(len(hostsdata)):  # get timestamps for labels
                hdata = hostsdata[ihost]
                nodename = hdata['nodename']
                if nodename not in charts.keys():
                    charts[nodename] = {
                        "sysname": hdata['sysname'],
                        "release": hdata['release'],
                        "machine": hdata['machine'],
                        "number-of-cpus": hdata['number-of-cpus'],
                        "timezone": hdata['timezone'],
                        "xlabels": [],
                        "activities": {},
                        "events": {'Restart': []}
                        }
                for istats in range(len(hdata['restarts'])):
                    for r in hdata['restarts'][istats].values():
                        date = f"{r['date']} {r['time']}"
                        if util.in_date_range(args, date):
                            charts[nodename]['xlabels'].append(date)
                            charts[nodename]['events']['Restart'].append(
                                {'date': date,
                                 'description': ('Restart events are included in sar files.'
                                                 + ' Check <b>sarcharts --help</b> to add your custom events and metrics.')
                                 })
                for istats in range(len(hdata['statistics'])):
                    for act, adata in hdata['statistics'][istats].items():
                        pbi += 1
                        pb.print_bar(pbi, f"data {act}.")
                        if act == "timestamp":
                            date = f"{adata['date']} {adata['time']}"
                            if not util.in_date_range(args, date):
                                continue
                            if (date not in charts[nodename]['xlabels']
                                    and util.in_date_range(args, date)):
                                charts[nodename]['xlabels'].append(date)
                            linedet = f"{hdata['nodename']};{adata['interval']};{date}"
                        else:
                            if not util.in_date_range(args, date):
                                continue
                            if isinstance(adata, list):
                                if act not in charts[nodename]['activities'].keys():
                                    line = linehead
                                    for h in adata[0].keys():
                                        line += f";{str(h)}"
                                    charts[nodename]['activities'][act] = {
                                        "content": [line.split(";")],
                                        "multiple": True
                                        }
                                for d in adata:
                                    line = linedet
                                    for v in d.values():
                                        line += f";{str(v)}"
                                    charts[nodename]['activities'][act][
                                        'content'].append(line.split(";"))
                            elif isinstance(adata, dict):
                                d = adata[list(adata.keys())[1]]
                                if isinstance(d, float) or isinstance(d, int):
                                    if (act not in charts[nodename][
                                            'activities'].keys()):
                                        line = linehead
                                        for h in adata.keys():
                                            line += f";{str(h)}"
                                        charts[nodename]['activities'][act] = {
                                            "content": [line.split(";")],
                                            "multiple": False
                                            }
                                    line = linedet
                                    for v in adata.values():
                                        line += f";{str(v)}"
                                    charts[nodename]['activities'][act][
                                        'content'].append(line.split(";"))
                                else:
                                    for subact, subdata in adata.items():
                                        nact = f"{act}_{subact}"
                                        if isinstance(subdata, list):
                                            if (nact not in charts[nodename][
                                                    'activities'].keys()):
                                                line = linehead
                                                for h in subdata[0].keys():
                                                    line += f";{str(h)}"
                                                charts[nodename]['activities'][nact] = {
                                                    "content": [line.split(";")],
                                                    "multiple": True
                                                    }
                                            for sv in subdata:
                                                line = linedet
                                                for v in sv.values():
                                                    line += f";{str(v)}"
                                                charts[nodename]['activities'][nact]['content'].append(line.split(";"))                                        
                                        elif isinstance(subdata, dict):
                                            if nact not in charts[nodename]['activities'].keys():
                                                line = linehead
                                                for h in subdata.keys():
                                                    line += f";{str(h)}"
                                                charts[nodename]['activities'][nact] = {
                                                    "content": [line.split(";")],
                                                    "multiple": False
                                                    }
                                            line = linedet
                                            for v in subdata.values():
                                                line += f";{str(v)}"
                                            charts[nodename]['activities'][nact]['content'].append(line.split(";"))                                        

                                        elif isinstance(subdata, float) or isinstance(subdata, int):
                                            if (nact not in charts[nodename][
                                                    'activities'].keys()):
                                                line = f"{linehead};{nact}"
                                                charts[nodename]['activities'][nact] = {
                                                    "content": [line.split(";")],
                                                    "multiple": False
                                                    }
                                            line = f"{linedet};{str(subdata)}"
                                            charts[nodename]['activities'][nact][
                                                'content'].append(line.split(";"))

        pb.finish("  Get data.")
        # write csv files
        for nodename, nodecharts in charts.items():
            for activity, csvdata in charts[nodename]['activities'].items():
                with open(f"{args.outputpath}/csvfiles/{nodename}_{activity}.csv", "w") as f:
                    # workaround for io.cs headers (and maybe other activities)
                    n = len(csvdata['content'][1])
                    f.write(";".join(csvdata['content'][0][:n]) + "\n")
                    csvdata['content'].pop(0)
                    csvdata['content'].sort(key=lambda x: x[2])
                    for line in csvdata['content']:
                        f.write(";".join(line) + "\n")

        # build the chartjs dict
        for nodename, nodecharts in charts.items():
            for activity, csvdata in nodecharts['activities'].items():
                csvfile = f"{args.outputpath}/csvfiles/{nodename}_{activity}.csv"
                with open(csvfile) as f:
                    charts[nodename]['activities'][activity]['datasets'] = {}
                    # set the first data field
                    datastart = 4 if csvdata['multiple'] else 3
                    # get headers from first line
                    line = f.readline().strip()
                    headers = line.split(";")[datastart:]
                    # get first stats date
                    pos = f.tell()
                    line = f.readline().split(";")
                    # seek file to first stats line
                    f.seek(pos)
                    for line in f:
                        fields = line.strip().split(";")
                        # insert a fake item on non multiple nodecharts
                        item = fields[3] if csvdata['multiple'] else ""
                        if item not in charts[nodename]['activities'][activity]['datasets'].keys():
                            charts[nodename]['activities'][activity]['datasets'][item] = []
                            for h in headers:
                                charts[nodename]['activities'][activity]['datasets'][item].append({
                                    "label": h,
                                    "values": []
                                    })
                        for f in range(len(fields[datastart:])):
                            charts[nodename]['activities'][activity]['datasets'][item][f]['values'].append({
                                    'x': fields[2],
                                    'y': fields[f+datastart]
                                    })
        return charts



import os
import setuptools


def read(fname):
    return open(
        os.path.join(os.path.dirname(__file__), fname), encoding="utf-8"
        ).read()


setuptools.setup(
    name='sarcharts',
    version='v0.3.2',
    setup_requires=['Jinja2'],
    scripts=['sarcharts/bin/__init__.py'],
    entry_points={
        'console_scripts': ['sarcharts=sarcharts.bin:main']
        },
    packages=setuptools.find_packages(),
    package_data={
        'sarcharts.conf': ['*'],
        'sarcharts.html.css': ['*'],
        'sarcharts.html.img': ['*'],
        'sarcharts.html.js': ['*'],
        'sarcharts.templates': ['*'],
    },
    license='GPLv3',
    author='Pablo Fernández Rodríguez',
    url='https://github.com/pafernanr/sarcharts',
    keywords='sysstat sar sadf',
    description="""
        SarCharts gets sysstat files from provided sarfilespaths
        and generates dynamic HTML Charts.""",
    long_description_content_type='text/markdown',
    long_description=read("README.md"),
    classifiers=[
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    )



import datetime
import os
import sys

import fnmatch
from pathlib import Path
import subprocess


# 'sar' default format on first place for performance
valid_date_formats = ["%Y-%m-%d %H:%M:%S",
                      "%Y-%m-%dT%H:%M:%S%z",
                      "%Y-%m-%dT%H:%M:%S",
                      "%Y-%m-%d %H:%M",
                      "%Y-%m-%d %H",
                      "%Y-%m-%d",
                      "%Y %b %d %H:%M:%S"
                      # 14/May/2024:14:49:20 +0200 # apache
                      # May 14 14:42:56 # messages
                      # Apr 21 13:01:01 # cron
                      ]


def debug(args, sev, msg):
    if args.quiet:
        return
    C = {
        'I': '\033[0;34m',
        'D': '\033[01;36m',
        '': '\033[0;32m',
        'W': '\033[93m',
        'E': '\033[0;31m',
        'RESET': '\033[0m'
        }
    levels = {'D': 0,
              'I': 1,
              'W': 2,
              'E': 3,
              '': 4
              }
    if sev == "":
        print(f"  {str(msg)}")
    elif levels[sev] >= levels[args.debug]:
        print(f"{C[sev]}[{sev}]{C['RESET']} {str(msg)}")
    if sev == 'E':
        sys.exit(1)


def exec_command(args, cmd):
    debug(args, "D", "execcommand: " + cmd)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    stdout = str(stdout.decode(encoding="utf-8", errors="ignore"))
    stderr = str(stderr.decode(encoding="utf-8", errors="ignore"))
    # if stderr != "":
    #    print(cmd + "\n" + stderr)
    #    sys.exit(1)
    return [stdout, stderr]


def get_filelist(filepaths):
    files = []
    for path in filepaths:
        # path is a folder
        if Path(path).is_dir():
            for f in os.listdir(path):
                if fnmatch.fnmatch(f, 'sa??'):
                    files.append(f"{path}/{str(f)}")
        # path is a file
        elif Path(path).is_file():
            files.append(str(path))
    return files


def sortfiles_by_mtime(files):
    details = {}
    for f in files:
        mtime = os.path.getmtime(f)
        details[mtime] = f
    # 'sorted' returns a 'set' hence no duplicates
    # but it still needs to be converted to list
    return list(dict(sorted(details.items())).values())


def valid_date(args, d):
    valid = valid_date_formats
    for v in valid:
        try:
            return datetime.datetime.strptime(d, v)
        except ValueError:
            continue
    debug(args, 'E', f"not a valid date: {d!r}. Valid formats: {str(valid)}")


def in_date_range(args, d):
    d = valid_date(args, d).timestamp()
    # use timestamps to avoid compare offset-naive and offset-aware datetimes
    if d >= args.fromdate.timestamp() and d <= args.todate.timestamp():
        return True
    else:
        return False



#!/usr/bin/python3
import os
import sys

sys.path.insert(0, "/usr/lib/tools/")

try:
    sys.path.insert(0, os.getcwd())
    from sarcharts import SarCharts
except KeyboardInterrupt as exc:
    raise SystemExit() from exc


def main():
    SarCharts().main()
    os._exit(0)


if __name__ == '__main__':
    main()
